/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <netinet/in.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <rmp440/rmp440.h>

enum rmp440_conf_item {
	RMP_CONFIG_MAGIC,
	RMP_CONFIG_VIDEO_SERVER_ONLINE,
	RMP_CONFIG_VIDEO_SERVER_IP,
	RMP_CONFIG_VIDEO_USERNAME,
	RMP_CONFIG_VIDEO_PASSWORD,
	RMP_CONFIG_OCU_CONTROLLER_TYPE,
	RMP_CONFIG_OCU_COMM_INTERFACE,
	RMP_CONFIG_MAX_VEL,
	RMP_CONFIG_MAX_ACCEL,
	RMP_CONFIG_MAX_DECEL,
	RMP_CONFIG_MAX_DTZ_DECEL,
	RMP_CONFIG_COASTDOWN_DECEL,
	RMP_CONFIG_MAX_YAW_RATE,
	RMP_CONFIG_MAX_YAW_ACCEL,
	RMP_CONFIG_TIRE_DIAMETER,
	RMP_CONFIG_WHEEL_BASE_LENGTH,
	RMP_CONFIG_WHEEL_TRACK_WIDTH,
	RMP_CONFIG_GEARBOX_RATIO,
	RMP_CONFIG_CONFIG_BITMAP,
	RMP_CONFIG_IP_ADDRESS,
	RMP_CONFIG_IP_PORT,
	RMP_CONFIG_IP_SUBNET_MASK,
	RMP_CONFIG_IP_GATEWAY,
	RMP_CONFIG_FEEDBACK_BITMAP_1,
	RMP_CONFIG_FEEDBACK_BITMAP_2,
	RMP_CONFIG_FEEDBACK_BITMAP_3,
	RMP_CONFIG_FEEDBACK_BITMAP_4,
	RMP_CONFIG_CONFIG_LOADED,
};

static struct rmp_440_conf_str {
	char *tag;
	enum rmp440_conf_item item;
} rmp440_conf_data[] = {
	{ "RMP_DEMO_OCU_CONFIGURATION",
	  RMP_CONFIG_MAGIC  },
	{ "Video Server Online:",
	  RMP_CONFIG_VIDEO_SERVER_ONLINE },
	{ "Video Server IP Address:",
	  RMP_CONFIG_VIDEO_SERVER_IP, },
	{ "Video Server Username:",
	  RMP_CONFIG_VIDEO_USERNAME, },
	{ "Video Server Password:",
	  RMP_CONFIG_VIDEO_PASSWORD, },
	{ "OCU Contoller Type:",
	  RMP_CONFIG_OCU_CONTROLLER_TYPE, },
	{ "OCU Comm Interface:",
	  RMP_CONFIG_OCU_COMM_INTERFACE, },
	{ "Max Velocity (m/s):",
	  RMP_CONFIG_MAX_VEL, },
	{ "Max Accel (m/s^2):",
	  RMP_CONFIG_MAX_ACCEL, },
	{ "Max Decel (m/s^2):",
	  RMP_CONFIG_MAX_DECEL, },
	{ "Max DTZ Decel (m/s^2):",
	  RMP_CONFIG_MAX_DTZ_DECEL, },
	{ "Coastdown Decel (m/s^2):",
	  RMP_CONFIG_COASTDOWN_DECEL, },
	{ "Max Yaw Rate (rad/s):",
	  RMP_CONFIG_MAX_YAW_RATE, },
	{ "Max Yaw Accel (rad/s^2):",
	  RMP_CONFIG_MAX_YAW_ACCEL, },
	{ "Tire Diameter (m):",
	  RMP_CONFIG_TIRE_DIAMETER, },
	{ "Wheel Base Length (m):",
	  RMP_CONFIG_WHEEL_BASE_LENGTH, },
	{ "Wheel Track Width (m):",
	  RMP_CONFIG_WHEEL_TRACK_WIDTH, },
	{ "Gearbox Ratio:",
	  RMP_CONFIG_GEARBOX_RATIO, },
	{ "Config Bitmap:",
	  RMP_CONFIG_CONFIG_BITMAP, },
	{ "RMP IP Address:",
	  RMP_CONFIG_IP_ADDRESS, },
	{ "RMP Port:",
	  RMP_CONFIG_IP_PORT, },
	{ "RMP Subnet Mask:",
	  RMP_CONFIG_IP_SUBNET_MASK, },
	{ "RMP Gateway:",
	  RMP_CONFIG_IP_GATEWAY, },
	{ "RMP User FB Bitmap 1:",
	  RMP_CONFIG_FEEDBACK_BITMAP_1, },
	{ "RMP User FB Bitmap 2:",
	  RMP_CONFIG_FEEDBACK_BITMAP_2, },
	{ "RMP User FB Bitmap 3:",
	  RMP_CONFIG_FEEDBACK_BITMAP_3, },
	{ "RMP User FB Bitmap 4:",
	  RMP_CONFIG_FEEDBACK_BITMAP_4, },
	{ "RMP Configuration Loaded:",
	  RMP_CONFIG_CONFIG_LOADED, }
};

#define nitems(a) (sizeof(a)/sizeof(a[0]))

static int
parseline(char *line, struct rmp440_config *conf)
{
	int i;
	char *val = NULL;
	char *end;

	for (i = 0; i < nitems(rmp440_conf_data); i++) {
		if (strncmp(line, rmp440_conf_data[i].tag,
			strlen(rmp440_conf_data[i].tag)) == 0) {
			val = line + strlen(rmp440_conf_data[i].tag);
			while (isspace(*val))
				val++;
			break;
		}
	}
	if (val == NULL)
		return -1;

	switch (rmp440_conf_data[i].item) {
	case RMP_CONFIG_MAGIC:
		break;
	case RMP_CONFIG_VIDEO_SERVER_ONLINE:
		conf->video_server_online = strtol(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_VIDEO_SERVER_IP:
		if (inet_aton(val, &conf->video_server_addr) == 0)
			return -1;
		break;
	case RMP_CONFIG_VIDEO_USERNAME:
		conf->video_server_username = strdup(strtok(val, "\n"));
		break;
	case RMP_CONFIG_VIDEO_PASSWORD:
		conf->video_server_password = strdup(strtok(val, "\n"));
		break;
	case RMP_CONFIG_OCU_CONTROLLER_TYPE:
		conf->ocu_controller_type = strdup(strtok(val, "\n"));
		break;
	case RMP_CONFIG_OCU_COMM_INTERFACE:
		end = strtok(val, "\n");
		if (strcmp(end, "Ethernet") == 0)
			conf->ocu_comm_interface = RMP440_INTERFACE_ETHERNET;
		else if (strcmp(end, "USB") == 0)
			conf->ocu_comm_interface = RMP440_INTERFACE_USB;
		else if (strcmp(end, "CAN") == 0)
			conf->ocu_comm_interface = RMP440_INTERFACE_CAN;
		else
			return -1;
		break;
	case RMP_CONFIG_MAX_VEL:
		conf->max_vel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_MAX_ACCEL:
		conf->max_accel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_MAX_DECEL:
		conf->max_decel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_MAX_DTZ_DECEL:
		conf->max_dtz_decel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_COASTDOWN_DECEL:
		conf->coastdown_decel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_MAX_YAW_RATE:
		conf->max_yaw_rate = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_MAX_YAW_ACCEL:
		conf->max_yaw_accel = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_TIRE_DIAMETER:
		conf->tire_diameter = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_WHEEL_BASE_LENGTH:
		conf->wheel_base_length = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_WHEEL_TRACK_WIDTH:
		conf->wheel_track_width = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_GEARBOX_RATIO:
		conf->gearbox_ratio = strtod(val, &end);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_CONFIG_BITMAP:
		conf->config_bitmap = strtoul(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_IP_ADDRESS:
		if (inet_aton(val, &conf->ip_address) == 0)
			return -1;
		break;
	case RMP_CONFIG_IP_PORT:
		conf->port = strtol(val, &end, 0);
		if (end == val || conf->port > UINT16_MAX)
			return -1;
		break;
	case RMP_CONFIG_IP_SUBNET_MASK:
		if (inet_aton(val, &conf->ip_netmask) == 0)
			return -1;
		break;
	case RMP_CONFIG_IP_GATEWAY:
		if (inet_aton(val, &conf->ip_gateway) == 0)
			return -1;
		break;
	case RMP_CONFIG_FEEDBACK_BITMAP_1:
		conf->feedback_bitmap[0] = strtoul(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_FEEDBACK_BITMAP_2:
		conf->feedback_bitmap[1] = strtoul(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_FEEDBACK_BITMAP_3:
		conf->feedback_bitmap[2] = strtoul(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_FEEDBACK_BITMAP_4:
		conf->feedback_bitmap[3] = strtoul(val, &end, 0);
		if (end == val)
			return -1;
		break;
	case RMP_CONFIG_CONFIG_LOADED:
		if (strncmp(val, "True", 4) == 0)
			conf->loaded = 1;
		else
			conf->loaded = 0;
		break;
	}
	return 0;
}

/**
 * Parse a Segway OCU configuration file
 */
struct rmp440_config *
rmp440ParseConfigFile(const char *filename)
{
	FILE *f;
	char *line = NULL;
	int n = 1;
	size_t len = 0;
	struct rmp440_config *config;

	config = (struct rmp440_config *)malloc(sizeof(struct rmp440_config));
	if (config == NULL)
		return NULL;

	f = fopen(filename, "r");
	if (f == NULL) {
		free(config);
		return NULL;
	}

	while (getline(&line, &len, f) != -1) {
		if (parseline(line, config) == -1) {
			fprintf(stderr, "%s:%d: syntax error\n", filename, n);
			free(config);
			config = NULL;
			break;
		}
		n++;
	}
	free(line);
	fclose(f);
	return config;
}

/**
 * Print the current configuration
 */
void
rmp440PrintConfig(const struct rmp440_config *conf, FILE *out)
{
	int i;
	for (i = 0; i < nitems(rmp440_conf_data); i++) {
		switch(rmp440_conf_data[i].item) {
		case RMP_CONFIG_MAGIC:
			fprintf(out, "%s\n", rmp440_conf_data[i].tag);
			break;
		case RMP_CONFIG_VIDEO_SERVER_ONLINE:
			fprintf(out, "%s %d\n", rmp440_conf_data[i].tag,
			    conf->video_server_online);
			break;
		case RMP_CONFIG_VIDEO_SERVER_IP:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    inet_ntoa(conf->video_server_addr));
			break;
		case RMP_CONFIG_VIDEO_USERNAME:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    conf->video_server_username);
			break;
		case RMP_CONFIG_VIDEO_PASSWORD:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    conf->video_server_password);
			break;
		case RMP_CONFIG_OCU_CONTROLLER_TYPE:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    conf->ocu_controller_type);
			break;
		case RMP_CONFIG_OCU_COMM_INTERFACE:
			fprintf(out, "%s ", rmp440_conf_data[i].tag);
			switch (conf->ocu_comm_interface) {
			case RMP440_INTERFACE_ETHERNET:
				printf("Ethernet");
				break;
			case RMP440_INTERFACE_USB:
				printf("USB");
				break;
			case RMP440_INTERFACE_CAN:
				printf("CAN");
				break;
			}
			printf("\n");
			break;
		case RMP_CONFIG_MAX_VEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_vel);
			break;
		case RMP_CONFIG_MAX_ACCEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_accel);
			break;
		case RMP_CONFIG_MAX_DECEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_decel);
			break;
		case RMP_CONFIG_MAX_DTZ_DECEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_dtz_decel);
			break;
		case RMP_CONFIG_COASTDOWN_DECEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->coastdown_decel);
			break;
		case RMP_CONFIG_MAX_YAW_RATE:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_yaw_rate);
			break;
		case RMP_CONFIG_MAX_YAW_ACCEL:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->max_yaw_accel);
			break;
		case RMP_CONFIG_TIRE_DIAMETER:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->tire_diameter);
			break;
		case RMP_CONFIG_WHEEL_BASE_LENGTH:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->wheel_base_length);
			break;
		case RMP_CONFIG_WHEEL_TRACK_WIDTH:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->wheel_track_width);
			break;
		case RMP_CONFIG_GEARBOX_RATIO:
			fprintf(out, "%s %.6lg\n", rmp440_conf_data[i].tag,
			    conf->gearbox_ratio);
			break;
		case RMP_CONFIG_CONFIG_BITMAP:
			fprintf(out, "%s %u\n", rmp440_conf_data[i].tag,
			    conf->config_bitmap);
			break;
		case RMP_CONFIG_IP_ADDRESS:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    inet_ntoa(conf->ip_address));
			break;
		case RMP_CONFIG_IP_PORT:
			fprintf(out, "%s %d\n", rmp440_conf_data[i].tag,
			    conf->port);
			break;
		case RMP_CONFIG_IP_SUBNET_MASK:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    inet_ntoa(conf->ip_netmask));
			break;
		case RMP_CONFIG_IP_GATEWAY:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    inet_ntoa(conf->ip_gateway));
			break;
		case RMP_CONFIG_FEEDBACK_BITMAP_1:
			fprintf(out, "%s %u\n", rmp440_conf_data[i].tag,
			    conf->feedback_bitmap[0]);
			break;
		case RMP_CONFIG_FEEDBACK_BITMAP_2:
			fprintf(out, "%s %u\n", rmp440_conf_data[i].tag,
			    conf->feedback_bitmap[1]);
			break;
		case RMP_CONFIG_FEEDBACK_BITMAP_3:
			fprintf(out, "%s %u\n", rmp440_conf_data[i].tag,
			    conf->feedback_bitmap[2]);
			break;
		case RMP_CONFIG_FEEDBACK_BITMAP_4:
			fprintf(out, "%s %u\n", rmp440_conf_data[i].tag,
			    conf->feedback_bitmap[3]);
			break;
		case RMP_CONFIG_CONFIG_LOADED:
			fprintf(out, "%s %s\n", rmp440_conf_data[i].tag,
			    conf->loaded ? "True" : "False");
			break;
		}
	} /* for */
}

/**
 * Initialize the feedback structure from a configuration
 */
void
rmp440FeedbackFromConfig(struct rmp440_feedback *fb,
    const struct rmp440_config *conf)
{
	int i;

	fb->fram_vel_limit = conf->max_vel;
	fb->fram_accel_limit = conf->max_accel;
	fb->fram_decel_limit = conf->max_decel;
	fb->fram_dtz_decel_limit = conf->max_dtz_decel;
	fb->fram_coastdown_decel = conf->coastdown_decel;
	fb->fram_yaw_rate_limit = conf->max_yaw_rate;
	fb->fram_yaw_accel_limit = conf->max_yaw_accel;
	fb->fram_wheel_base_length = conf->wheel_base_length;
	fb->fram_wheel_track_width = conf->wheel_track_width;
	fb->fram_transmission_ratio = conf->gearbox_ratio;
	fb->fram_config_bitmap = conf->config_bitmap;
	fb->fram_eth_ip_address =  conf->ip_address.s_addr;
	fb->fram_eth_port_number = conf->port;
	fb->fram_eth_subnet_mask = conf->ip_netmask.s_addr;
	fb->fram_eth_gateway = conf->ip_gateway.s_addr;
	for (i = 0; i < 4; i++)
		fb->user_feedback_bitmap[i] = conf->feedback_bitmap[i];
}
