/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <rmp440/rmp440.h>

#include "io.h"
#include "crc16.h"

/**
 * Init a feedback structure from an optional configuration file.
 *
 */
struct rmp440_feedback *
rmp440FeedbackInit(const char *filename)
{
	struct rmp440_feedback *fb;
	struct rmp440_config *config;
	int i;

	fb = (struct rmp440_feedback *)malloc(sizeof(struct rmp440_feedback));
	if (fb == NULL)
		return NULL;
	if (filename != NULL) {
		config = rmp440ParseConfigFile(filename);
		if (config == NULL) {
			fprintf(stderr,
			    "configuration file parsing error\n");
			free(fb);
			return NULL;
		} else {
			rmp440FeedbackFromConfig(fb, config);
			free(config);
		}
	} else {
		/* Init user feedback bitmaps to default values
		 * hoping they match the actual NVRAM values */
		for (i = 0; i < 3; i++) {
			fb->user_feedback_bitmap[i] = 0xffffffff;
		}
		fb->user_feedback_bitmap[3] = 0x0;
	}
	return fb;
}

static int
rmp440DecodeInt(struct rmp440_feedback *fb, uint32_t val, int idx)
{
	switch (idx) {
	case CCU_FSW_1_INDEX:
		fb->fault_status[0] = val;
		break;
	case CCU_FSW_2_INDEX:
		fb->fault_status[1] = val;
		break;
	case CCU_FSW_3_INDEX:
		fb->fault_status[2] = val;
		break;
	case CCU_FSW_4_INDEX:
		fb->fault_status[3] = val;
		break;
	case CCU_MCU_0_FSW_INDEX:
		fb->mcu_fault_status[0] = val;
		break;
	case CCU_MCU_1_FSW_INDEX:
		fb->mcu_fault_status[1] = val;
		break;
	case CCU_MCU_2_FSW_INDEX:
		fb->mcu_fault_status[2] = val;
		break;
	case CCU_MCU_3_FSW_INDEX:
		fb->mcu_fault_status[3] = val;
		break;
	case CCU_OPERATIONAL_STATE_INDEX:
		fb->operational_state = val;
		break;
	case CCU_DYNAMIC_RESPONSE_INDEX:
		fb->dynamic_response = val;
		break;
	case CCU_PSE_DATA_VALID_INDEX:
		fb->pse_data_is_valid = val;
		break;
	case CCU_AUX_BATT_ABB_STATUS_INDEX:
		fb->abb_system_status = val;
		break;
	case CCU_AUX_BATT_STATUS_INDEX:
		fb->abb_batt_status = val;
		break;
	case CCU_AUX_BATT_FAULTS_INDEX:
		fb->aux_batt_faults = val;
		break;
	case CCU_SP_SW_BUILD_ID_INDEX:
		fb->sp_sw_build_id = val;
		break;
	case CCU_UIP_SW_BUILD_ID_INDEX:
		fb->uip_sw_build_id = val;
		break;
	case CCU_FRAM_CFG_BITMAP_INDEX:
		fb->fram_config_bitmap = val;
		break;
	case CCU_FRAM_ETH_PORT_NUMBER_INDEX:
		fb->fram_eth_port_number = val;
		break;
	case CCU_FRAM_USER_FEEDBACK_1_INDEX:
		fb->user_feedback_bitmap[0] = val;
		break;
	case CCU_FRAM_USER_FEEDBACK_2_INDEX:
		fb->user_feedback_bitmap[1] = val;
		break;
	case CCU_FRAM_USER_FEEDBACK_3_INDEX:
		fb->user_feedback_bitmap[2] = val;
		break;
	case CCU_FRAM_USER_FEEDBACK_4_INDEX:
		fb->user_feedback_bitmap[3] = val;
		break;
	default:
		fprintf(stderr, "rmp440DecodeInt: invalid index %d\n", idx);
		return -1;
	}
	return 0;
}

/*
 * Decode IPv4 addresses stored as 32 bit values
 *
 * XXX should use sockaddr_in type and properly convert them to/from
 * XXX network order if they need to be used for real.
 */
static int
rmp440DecodeIP(struct rmp440_feedback *fb, uint32_t val, int idx)
{
	switch (idx) {
	case CCU_FRAM_ETH_IP_ADDRESS_INDEX:
		fb->fram_eth_ip_address = val;
		break;
	case CCU_FRAM_ETH_SUBNET_MASK_INDEX:
		fb->fram_eth_subnet_mask = val;
		break;
	case CCU_FRAM_ETH_GATEWAY_INDEX:
		fb->fram_eth_gateway = val;
		break;
	default:
		fprintf(stderr, "rmp440DecodeIP: invalid index %d\n", idx);
		return -1;
	}
	return 0;
}

static int
rmp440DecodeFloat(struct rmp440_feedback *fb, float val, int idx)
{
	switch (idx) {
	case CCU_FRAME_COUNT_INDEX:
		fb->frame_count = val;
		break;
	case CCU_MIN_PROP_SOC_INDEX:
		fb->min_propulsion_batt_soc = val;
		break;
	case CCU_AUX_BATT_SOC_INDEX:
		fb->aux_batt_soc = val;
		break;
	case CCU_INERTIAL_X_ACC_INDEX:
		fb->inertial_x_acc = val;
		break;
	case CCU_INERTIAL_Y_ACC_INDEX:
		fb->inertial_y_acc = val;
		break;
	case CCU_INERTIAL_X_RATE_INDEX:
		fb->inertial_x_rate = val;
		break;
	case CCU_INERTIAL_Y_RATE_INDEX:
		fb->inertial_y_rate = val;
		break;
	case CCU_INERTIAL_Z_RATE_INDEX:
		fb->inertial_z_rate = val;
		break;
	case CCU_PSE_PITCH_INDEX:
		fb->pse_pitch = val;
		break;
	case CCU_PSE_PITCH_RATE_INDEX:
		fb->pse_pitch_rate = val;
		break;
	case CCU_PSE_ROLL_INDEX:
		fb->pse_roll = val;
		break;
	case CCU_PSE_ROLL_RATE_INDEX:
		fb->pse_roll_rate = val;
		break;
	case CCU_PSE_YAW_RATE_INDEX:
		fb->pse_yaw_rate = val;
		break;
	case CCU_YAW_RATE_LIMIT_INDEX:
		fb->yaw_rate_limit = val;
		break;
	case CCU_VELOCITY_LIMIT_INDEX:
		fb->vel_limit = val;
		break;
	case CCU_LINEAR_ACCEL_INDEX:
		fb->linear_accel = val;
		break;
	case CCU_LINEAR_VELOCITY_INDEX:
		fb->linear_vel = val;
		break;
	case CCU_DIFF_WHEEL_VELOCITY_INDEX:
		fb->differential_wheel_vel = val;
		break;
	case CCU_RF_WHEEL_VEL_INDEX:
		fb->right_front_vel = val;
		break;
	case CCU_LF_WHEEL_VEL_INDEX:
		fb->left_front_vel = val;
		break;
	case CCU_RR_WHEEL_VEL_INDEX:
		fb->right_rear_vel = val;
		break;
	case CCU_LR_WHEEL_VEL_INDEX:
		fb->left_rear_vel = val;
		break;
	case CCU_RF_WHEEL_POS_INDEX:
		fb->right_front_pos = val;
		break;
	case CCU_LF_WHEEL_POS_INDEX:
		fb->left_front_pos = val;
		break;
	case CCU_RR_WHEEL_POS_INDEX:
		fb->right_rear_pos = val;
		break;
	case CCU_LR_WHEEL_POS_INDEX:
		fb->left_rear_pos = val;
		break;
	case CCU_LINEAR_POS_INDEX:
		fb->linear_pos = val;
		break;
	case CCU_RF_MOTOR_CURRENT_INDEX:
		fb->right_front_current = val;
		break;
	case CCU_LF_MOTOR_CURRENT_INDEX:
		fb->left_front_current = val;
		break;
	case CCU_RR_MOTOR_CURRENT_INDEX:
		fb->right_rear_current = val;
		break;
	case CCU_LR_MOTOR_CURRENT_INDEX:
		fb->left_rear_current = val;
		break;
	case CCU_MAX_MOTOR_CURRENT_INDEX:
		fb->max_motor_current = val;
		break;
	case CCU_RF_MOTOR_CURRENT_LIMIT_INDEX:
		fb->right_front_current_limit = val;
		break;
	case CCU_LF_MOTOR_CURRENT_LIMIT_INDEX:
		fb->left_front_current_limit = val;
		break;
	case CCU_RR_MOTOR_CURRENT_LIMIT_INDEX:
		fb->right_rear_current_limit = val;
		break;
	case CCU_LR_MOTOR_CURRENT_LIMIT_INDEX:
		fb->left_rear_current_limit = val;
		break;
	case CCU_MIN_CURRENT_LIMIT_INDEX:
		fb->min_motor_current_limit = val;
		break;
	case CCU_FPB_1_SOC_INDEX:
		fb->front_base_batt_1_soc = val;
		break;
	case CCU_FPB_2_SOC_INDEX:
		fb->front_base_batt_2_soc = val;
		break;
	case CCU_RPB_1_SOC_INDEX:
		fb->rear_base_batt_1_soc = val;
		break;
	case CCU_RPB_2_SOC_INDEX:
		fb->rear_base_batt_2_soc = val;
		break;
	case CCU_FPB_1_TEMP_INDEX:
		fb->front_base_batt_1_temp = val;
		break;
	case CCU_FPB_2_TEMP_INDEX:
		fb->front_base_batt_2_temp = val;
		break;
	case CCU_RPB_1_TEMP_INDEX:
		fb->rear_base_batt_1_temp = val;
		break;
	case CCU_RPB_2_TEMP_INDEX:
		fb->rear_base_batt_2_temp = val;
		break;
	case CCU_VEL_TARGET_INDEX:
		fb->vel_target = val;
		break;
	case CCU_YAW_TARGET_INDEX:
		fb->yaw_rate_target = val;
		break;
	case CCU_ANGLE_TARGET_INDEX:
		fb->angle_target = val;
		break;
	case CCU_AUX_BATT_VOLTAGE_INDEX:
		fb->aux_batt_voltage = val;
		break;
	case CCU_AUX_BATT_CURRENT_INDEX:
		fb->aux_batt_current = val;
		break;
	case CCU_AUX_BATT_TEMP_INDEX:
		fb->aux_batt_temp = val;
		break;
	case CCU_7P2V_BATTERY_VOLTAGE_INDEX:
		fb->ccu_7p2_battery_voltage = val;
		break;
	case CCU_MCU_0_INST_POWER_INDEX:
		fb->mcu_inst_power[0] = val;
		break;
	case CCU_MCU_1_INST_POWER_INDEX:
		fb->mcu_inst_power[1] = val;
		break;
	case CCU_MCU_2_INST_POWER_INDEX:
		fb->mcu_inst_power[2] = val;
		break;
	case CCU_MCU_3_INST_POWER_INDEX:
		fb->mcu_inst_power[3] = val;
		break;
	case CCU_MCU_0_TOTAL_ENERGY_INDEX:
		fb->mcu_total_energy[0] = val;
		break;
	case CCU_MCU_1_TOTAL_ENERGY_INDEX:
		fb->mcu_total_energy[1] = val;
		break;
	case CCU_MCU_2_TOTAL_ENERGY_INDEX:
		fb->mcu_total_energy[2] = val;
		break;
	case CCU_MCU_3_TOTAL_ENERGY_INDEX:
		fb->mcu_total_energy[3] = val;
		break;
	case CCU_FRAM_VEL_LIMIT_INDEX:
		fb->fram_vel_limit = val;
		break;
	case CCU_FRAM_ACCEL_LIMIT_INDEX:
		fb->fram_accel_limit = val;
		break;
	case CCU_FRAM_DECEL_LIMIT_INDEX:
		fb->fram_decel_limit = val;
		break;
	case CCU_FRAM_MAX_DTZ_DECEL_INDEX:
		fb->fram_dtz_decel_limit = val;
		break;
	case CCU_FRAM_COASTDOWN_ACCEL_INDEX:
		fb->fram_coastdown_decel = val;
		break;
	case CCU_FRAM_YAW_RATE_LIMIT_INDEX:
		fb->fram_yaw_rate_limit = val;
		break;
	case CCU_FRAM_YAW_ACCEL_LIMIT_INDEX:
		fb->fram_yaw_accel_limit = val;
		break;
	case CCU_FRAM_TIRE_DIAMETER_INDEX:
		fb->fram_tire_diameter = val;
		break;
	case CCU_FRAM_WHEEL_BASE_INDEX:
		fb->fram_wheel_base_length = val;
		break;
	case CCU_FRAM_WHEEL_TRACK_WIDTH_INDEX:
		fb->fram_wheel_track_width = val;
		break;
	case CCU_FRAM_TRANSMISSION_RATIO_INDEX:
		fb->fram_transmission_ratio = val;
		break;
	default:
		fprintf(stderr, "rmp440DecodeFloat: invalid index %d\n",
		    idx);
		return -1;
	}
	return 0;
}

/*
 * Masks indicating the type of returned values. Default == unsigned int
 */
static uint32_t rmp_float_feedback_mask[4] = {
	0xFF7FF900, 0x3FFFFFFF, 0x007FFFF2, 0x00000000
};

static uint32_t rmp_ip_feedback_mask[4] = {
	0x0, 0x0, 0x0D000000, 0x0
};

static int
rmp440DecodeVal(struct rmp440_feedback *fb, uint8_t *val, int idx)
{
	union {
		uint32_t i;
		float f;
	} u;
	int i, mask;

	u.i = ((val[0] << 24) & 0xff000000) |
	    ((val[1] << 16) & 0x00ff0000) |
	    ((val[2] << 8) & 0x0000ff00) |
	    (val[3] & 0x000000ff);

	i = idx / 32;
	mask = 1 << (idx % 32);

	if ((rmp_float_feedback_mask[i] & mask) == mask)
		return rmp440DecodeFloat(fb, u.f, idx);
	else if ((rmp_ip_feedback_mask[i] & mask) == mask)
		return rmp440DecodeIP(fb, u.i, idx);
	else
		return rmp440DecodeInt(fb, u.i, idx);
}

int
rmp440Decode(struct rmp440_feedback *fb, uint8_t *buf, size_t len)
{
	int i, j;
	uint32_t mask;

	if (!crc_is_valid(buf, len)) {
		fprintf(stderr, "rmp440Decode: bad CRC\n");
		return -1;
	}

	/* For each possible feedback, if the corresponding bit is present
	 * in the user mask, then read and decode the value */
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 32; j++) {
			mask = 1 << j;
			if ((fb->user_feedback_bitmap[i] & mask) == mask) {
				if (rmp440DecodeVal(fb, buf, i*32+j) < 0)
					return -1;
				buf += sizeof(uint32_t);
			}
		}
	}
	return 0;
}

static const char *rmp440OperationalState[] = {
	"SYS_INIT0",
	"SYS_INIT1",
	"SYS_INIT2",
	"STANDBY",
	"TRACTOR",
	"DISABLE"
};
#define RMP_NUM_OPERATIONAL_STATES 6

const char *
rmp440DecodeOperationalState(uint32_t os)
{
	if (os >= RMP_NUM_OPERATIONAL_STATES)
		return "UNKNOWN";
	else
		return rmp440OperationalState[os];
}


/*
 * Number of bits in a u32
 *
 * http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
 */
static size_t
popcount(uint32_t v)
{
	v = v - ((v >> 1) & 0x55555555); // reuse input as temporary
	v = (v & 0x33333333) + ((v >> 2) & 0x33333333);	// temp
	return (((v + (v >> 4)) & 0xF0F0F0F) * 0x1010101) >> 24; // count

}

/*
 * rmp440ExpectedBytes - compute the size of a user feedback message
 *
 * number of expected value * sizeof u32 + sizeof crc
 */
static size_t
rmp440ExpectedBytes(const struct rmp440_feedback *fb)
{
	int i, n = 0;

	for (i = 0; i < 4; i++)
		n += popcount(fb->user_feedback_bitmap[i]);
	return n*sizeof(uint32_t) + sizeof(uint32_t);
}

int
rmp440ReceiveAndDecode(const struct rmp440_io *rmp, struct rmp440_feedback *fb)
{
	ssize_t nread;
	int bad_frames = 0;
	uint8_t data[1024];

	nread = rmp440Receive(rmp, data, sizeof(data));
	if (nread != rmp440ExpectedBytes(fb)) {
		fprintf(stderr, "%s: short read %d (%d)\n", __func__,
		    (int)nread, (int)rmp440ExpectedBytes(fb));
		bad_frames++;
	} else {
		if (rmp440Decode(fb, data, nread) == -1)
			bad_frames++;
		else
			bad_frames = 0;
	}
	if (bad_frames > 10) {
		printf("rmpTask: Communication error\n");
		return -1;
	}
	return 0;
}
