/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <errno.h>
#include <netdb.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

#include "io.h"

static ssize_t rmpInetSend(const struct rmp440_io *, const uint8_t *);
static ssize_t rmpInetReceive(const struct rmp440_io *, uint8_t *, size_t);
static void rmpInetClose(const struct rmp440_io *);

struct rmp440_io *
rmpInetInit(const char *addr)
{
	struct rmp440_io *rmp;
	struct addrinfo hints, *res, *res0, *res1, *lcl;
	char *addr1, *port;
	int error, sock;

	/* Rough parsing of addr which should be host:port */
	addr1 = strdup(addr);
	if (addr1 == NULL)
		return NULL;
	port = strchr(addr1, ':');
	if (port != NULL) {
		*port = '\0';	/* terminate host part */
		port++;		/* skip to port 1st char */
	}
	/* Setup socket - IPv6 compatible */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family  = PF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	error = getaddrinfo(addr1, port, &hints, &res0);
	if (error) {
		fprintf(stderr, "%s: %s\n", addr1, gai_strerror(error));
		return NULL;
	}
	for (res = res0; res != NULL; res = res->ai_next) {
		sock = socket(res->ai_family, res->ai_socktype,
		    res->ai_protocol);
		if (sock != -1)
			break;
	}
	if (res == NULL) {
		perror("sock");
		freeaddrinfo(res0);
		return NULL;
	}
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = res->ai_family;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;
	error = getaddrinfo(NULL, port, &hints, &res1);
	if (error) {
		fprintf(stderr, "passive: %s\n", gai_strerror(error));
		return NULL;
	}
	for (lcl = res1; lcl != NULL; lcl = lcl->ai_next) {
		if (bind(sock, lcl->ai_addr, lcl->ai_addrlen) == 0)
		    break;
	}
	if (lcl == NULL) {
		perror("bind");
		freeaddrinfo(res0);
		freeaddrinfo(res1);
		return NULL;
	}
	if (connect(sock, res->ai_addr, res->ai_addrlen) == -1) {
		perror("connect");
		freeaddrinfo(res0);
		freeaddrinfo(res1);
		return NULL;
	}
	freeaddrinfo(res0);
	freeaddrinfo(res1);

	/* Allocate descriptor */
	rmp = (struct rmp440_io *)malloc(sizeof(struct rmp440_io));
	if (rmp == NULL) {
		close(sock);
		return NULL;
	}
	rmp->fd = sock;
	rmp->send = rmpInetSend;
	rmp->receive = rmpInetReceive;
	rmp->close = rmpInetClose;
	return rmp;
}

static ssize_t
rmpInetSend(const struct rmp440_io *rmp, const uint8_t *buffer)
{

	return write(rmp->fd, buffer, RMP_NUM_USB_ETH_BYTES);
}

static ssize_t
rmpInetReceive(const struct rmp440_io *rmp, uint8_t *buffer, size_t len)
{
	return read(rmp->fd, buffer, len);
}

static void
rmpInetClose(const struct rmp440_io *rmp)
{
	close(rmp->fd);
}
