/*
 * Copyright (c) 2003-2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

#include "io.h"

/* RMP440 usb-to serial adapter baud rate is fixed */
#define RMP440_BAUD_RATE 115200

static ssize_t rmpUsbSend(const struct rmp440_io *, const uint8_t *);
static ssize_t rmpUsbReceive(const struct rmp440_io *, uint8_t *, size_t);
static void rmpUsbClose(const struct rmp440_io *);
static int openSerial(const char *, int);

struct rmp440_io *
rmpUsbInit(const char *port)
{
	struct rmp440_io *rmp;

	rmp = (struct rmp440_io *)malloc(sizeof(struct rmp440_io));
	if (rmp == NULL)
		return NULL;

	rmp->fd = openSerial(port, RMP440_BAUD_RATE);
	if (rmp->fd == -1)
		return NULL;
	rmp->send = rmpUsbSend;
	rmp->receive = rmpUsbReceive;
	rmp->close = rmpUsbClose;
	return rmp;
}

static ssize_t
rmpUsbSend(const struct rmp440_io *rmp, const uint8_t *buffer)
{
	return write(rmp->fd, buffer, RMP_NUM_USB_ETH_BYTES);
}

static ssize_t
rmpUsbReceive(const struct rmp440_io *rmp, uint8_t *buffer, size_t len)
{
	usleep(50000);
	return read(rmp->fd, buffer, len);
}

static void
rmpUsbClose(const struct rmp440_io *rmp)
{
	close(rmp->fd);
}

/*
 * Open a serial port in raw mode with the given baud rate
 */
static int
openSerial(const char *device, int speed)
{
	int fd, flags;
	struct termios t;

	fd = open(device, O_RDWR|O_NONBLOCK);
	if (fd < 0) {
		fprintf(stderr, "%s:%d: %s: %s\n", __FILE__, __LINE__,
		    device, strerror(errno));
		return -1;
	}
	printf ("openSerial: open %s ok fd = %d\n", device, fd);
	tcgetattr(fd, &t);
	cfmakeraw(&t);

	switch (speed) {
	case 1200:
		cfsetspeed(&t, B1200);
		break;
	case 2400:
		cfsetspeed(&t, B2400);
		break;
	case 4800:
		cfsetspeed(&t, B4800);
		break;
	case 9600:
		cfsetspeed(&t, B9600);
		break;
	case 19200:
		cfsetspeed(&t, B19200);
		break;
	case 38400:
		cfsetspeed(&t, B38400);
		break;
	case 57600:
		cfsetspeed(&t, B57600);
		break;
	case 115200:
		cfsetspeed(&t, B115200);
		break;
	default:
		fprintf(stderr, "%s:%d: unknown speed %d\n", __FILE__,
		    __LINE__, speed);
		close(fd);
		return -1;
	}
	if (tcsetattr(fd, TCSAFLUSH, &t) < 0) {
		fprintf(stderr, "%s:%d: tcsetattr: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}
	flags = fcntl(fd, F_GETFL, 0);
	if (flags < 0) {
		fprintf(stderr, "%s:%d: F_GETFL: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}
	flags &= ~O_NONBLOCK;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		fprintf(stderr, "%s:%d: F_SETFL: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}

	return fd;
}
