/*
 * Copyright (c) 2009,2010,2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>

#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <linux/joystick.h>

#include "rmp440/rmp440.h"

#define  SIGN(x)             ((x)<0. ? -1. : 1.)

/* For joystick driver */
#define NAME_LENGTH 128

struct rmp440_io *rmp;
struct rmp440_feedback *fb;

unsigned char axes = 2;		/* number of joystick axes */
unsigned char buttons = 2;	/* number of joystick buttons */
int *axis;			/* array of joystick axis values */
int *button;			/* array of joystick button state */
float v, w;			/* linear & angular speed references */
int quit;			/* quit flag */

int version = 0x000800;
char name[NAME_LENGTH] = "Unknown";

pthread_mutex_t joymutex = PTHREAD_MUTEX_INITIALIZER;


/*----------------------------------------------------------------------*/
/**
 ** Init joystick
 **/
static int
joyinit(char *dev)
{
	int joy;

	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return -1;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);

	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16,
	       (version >> 8) & 0xff, version & 0xff);
	return joy;
}
/*----------------------------------------------------------------------*/

/**
 ** task reading data from the joystick and producing references;
 **/
void *
joyRead(void *arg)
{
	struct js_event js;
	struct pollfd pfd;
	int ret;
	int fd = (int)(intptr_t)arg;

	while (!quit) {
		pfd.fd = fd;
		pfd.events = POLLIN;
		pfd.revents = 0;
		ret = poll(&pfd, 1, 100);

		if (ret < 0) {
			if (errno != EINTR)
				fprintf(stderr, "poll error: %s\n",
					strerror(errno));
			continue;
		}
		if (ret == 0)
			/* nothing to read from joystick */
			continue;
		if ((pfd.revents & POLLIN) != POLLIN)
			continue;
		/* something to read from joystick */
		if (read(fd, &js, sizeof(struct js_event))
		    != sizeof(struct js_event)) {
			perror("\njstest: error reading");
			break;
		}
		switch(js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			button[js.number] = js.value;
			printf("button %d %d\n", js.number, js.value);
			break;
		case JS_EVENT_AXIS:
			axis[js.number] = js.value;
			printf("axis %d %d\n", js.number, js.value);
			break;
		}
		pthread_mutex_lock(&joymutex);
		/* Test for motion button */
		if (button[0]) {
			/* compute speed refs */
			v = -SIGN(axis[1]) * (axis[1]*axis[1])/(32768.0*32768.0);
			w = SIGN(axis[0]) * (axis[0]*axis[0])/(32768.0*32768.0);
			printf("\t\tmove %f %f\n", v, w);
		} else {
			/* Set refs to zero */
			printf("stop\n");
			v = 0;
			w = 0;
		}
		pthread_mutex_unlock(&joymutex);
	}
	return NULL;
}
/*----------------------------------------------------------------------*/

/**
 ** Task reading data from one RMP440 device.
 ** Frames are decoded and the device state is updated, but not used.
 **/
static void *
rmpTask(void *arg)
{
	struct timespec tv;

	while (!quit) {
		/* Test for quit button */
		if (button[3])
			quit = 1;

		tv.tv_sec = 0;
		tv.tv_nsec = 10000000; /* 10ms */
		nanosleep(&tv, NULL);

		switch (fb->operational_state) {
		case 0:
		case 1:
		case 2:
			rmp440CmdNone(rmp);
			break;
		case 3:/* standby */
			if (button[0])
				rmp440SetOperationalMode(rmp,
				    RMP_TRACTOR_REQUEST);
			else
				rmp440CmdNone(rmp);
			break;
		case 4: /* tractor */
			if (button[0])
				rmp440Motion(rmp, v, w);
			else
				rmp440SetOperationalMode(rmp,
				    RMP_STANDBY_REQUEST);
			break;
		default:
			rmp440CmdNone(rmp);
			break;

		}
		rmp440ReceiveAndDecode(rmp, fb);
	}
	return NULL;
}

/*----------------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
	pthread_t tid, jtid;
	int j;

	v = 0;
	w = 0;
	quit = 0;


	/* Connect to RMP device */
	fb = rmp440FeedbackInit(NULL);
	if (fb == NULL) {
		fprintf(stderr, "Cannot allocate memory\n");
		exit(2);
	}

	if (argc == 1)
		rmp = rmp440Init(RMP440_INTERFACE_ETHERNET,
		    "192.168.0.40:8080");
	else {
		if (argv[1][0] == '/') /* /dev/ttyACM0 */
			rmp = rmp440Init(RMP440_INTERFACE_USB,
			    argv[1]);
		else if (argv[1][0] >= '0' && argv[1][0] <= '9')
			/* ip address */
			rmp = rmp440Init(RMP440_INTERFACE_ETHERNET,
			    argv[1]);
	}
	if (rmp == NULL) {
		fprintf(stderr, "Cannot init rmp connection\n");
		exit(2);
	}
	printf("rmp440Init done\n");


	/* Connect to joystick */
	j = joyinit("/dev/input/js0");
	if (j < 0) {
		fprintf(stderr, "Error initializing joystick\n");
		exit(2);
	}
	printf("joyinit done\n");
	axis = calloc(axes, sizeof(int));
	button = calloc(buttons, sizeof(int));

	printf("use button 4 to quit\n"
	       "   button 1 + left stick to move\n");

	/* spawn RMP reader task */
	pthread_create(&tid, NULL, rmpTask, NULL);

	pthread_create(&jtid, NULL, joyRead, (void *)(intptr_t)j);

	printf("go\n");

	pthread_join(tid, NULL);
	pthread_join(jtid, NULL);

	/* Free resources */
	free(button);
	free(axis);
	close (j);

	return 0;
}
